import json
from collections.abc import Callable, Iterable, Sequence
from itertools import islice
from pathlib import Path

import pytest
from faker import Faker
from faker.typing import SeedType
from pydantic import BaseModel

from mex.artificial.constants import DEFAULT_LOCALE, DEFAULT_MODELS
from mex.artificial.helpers import (
    create_artificial_extracted_items,
    create_artificial_items_and_rule_sets,
    create_artificial_merged_items,
    create_faker,
    generate_artificial_extracted_items,
    generate_artificial_items_and_rule_sets,
    generate_artificial_merged_items,
    write_merged_items,
)
from mex.artificial.types import LocaleType

TEST_DATA_PATH = Path(__file__).parent / "test_data"


def test_faker() -> None:
    faker = create_faker(["en_US", "de_DE"], 42, 1)
    assert isinstance(faker, Faker)
    assert faker.locales == ["en_US", "de_DE"]
    assert faker.text_string() == "site"


@pytest.mark.parametrize(
    ("generator_func", "expected_file"),
    [
        (
            generate_artificial_extracted_items,
            "extracted_items.json",
        ),
        (
            generate_artificial_items_and_rule_sets,
            "extracted_items_and_rule_sets.json",
        ),
        (
            generate_artificial_merged_items,
            "merged_items.json",
        ),
    ],
    ids=["extracted_items", "items_and_rule_sets", "merged_items"],
)
def test_generate_artificial_items(
    generator_func: Callable[
        [LocaleType, SeedType, int, Sequence[str]], Iterable[BaseModel]
    ],
    expected_file: str,
) -> None:
    gen = generator_func(DEFAULT_LOCALE, 42, 5, DEFAULT_MODELS)
    items = list(islice(gen, 10))
    output = [item.model_dump(mode="json") for item in items]

    expected_path = TEST_DATA_PATH / expected_file
    with expected_path.open(encoding="utf-8") as fh:
        expected = json.load(fh)

    assert output == expected


@pytest.mark.parametrize(
    ("creator_func", "expected_file"),
    [
        (
            create_artificial_extracted_items,
            "extracted_items.json",
        ),
        (
            create_artificial_items_and_rule_sets,
            "extracted_items_and_rule_sets.json",
        ),
        (
            create_artificial_merged_items,
            "merged_items.json",
        ),
    ],
    ids=["extracted_items", "items_and_rule_sets", "merged_items"],
)
def test_create_artificial_items(
    creator_func: Callable[
        [LocaleType, SeedType, int, Sequence[str], int], list[BaseModel]
    ],
    expected_file: str,
) -> None:
    items = creator_func(DEFAULT_LOCALE, 42, 5, DEFAULT_MODELS, 10)
    output = [item.model_dump(mode="json") for item in items]

    expected_path = TEST_DATA_PATH / expected_file
    with expected_path.open(encoding="utf-8") as fh:
        expected = json.load(fh)

    assert output == expected


def test_write_merged_items(tmp_path: Path) -> None:
    gen = generate_artificial_merged_items(DEFAULT_LOCALE, 42, 5, DEFAULT_MODELS)
    write_merged_items(gen, 23, tmp_path)

    output_file = tmp_path / "publisher.ndjson"
    assert output_file.exists()

    with output_file.open(encoding="utf-8") as fh:
        lines = fh.readlines()

    assert len(lines) == 23
    for line in lines:
        item = json.loads(line)
        assert "identifier" in item
        assert "entityType" in item



import pytest
from faker import Faker

from mex.artificial.helpers import create_faker
from mex.common.models import BASE_MODEL_CLASSES

pytest_plugins = ("mex.common.testing.plugin",)


@pytest.fixture(name="faker")
def init_faker() -> Faker:
    """Return a fully configured faker instance."""
    return create_faker(["en_US"], 0, 5)


@pytest.fixture
def ids_by_type() -> dict[str, list[str]]:
    """Return a mapping of stemTypes to lists of dummy identifiers."""
    return {
        model.stemType: [
            f"{model.stemType}{str(i).zfill(21 - len(model.stemType))}"
            for i in range(1, 9)
        ]
        for model in BASE_MODEL_CLASSES
    }



import re
from collections import OrderedDict
from collections.abc import Callable, Collection, Mapping, Sequence
from datetime import datetime
from enum import Enum
from functools import partial
from typing import Any, cast, get_args, get_origin

from annotated_types import MaxLen, MinLen
from faker import Generator as FakerFactory
from faker.providers import BaseProvider as BaseFakerProvider
from faker.providers.internet import Provider as InternetFakerProvider
from faker.providers.python import Provider as PythonFakerProvider
from pydantic import ValidationError
from pydantic.fields import FieldInfo

from mex.artificial.constants import (
    DEFAULT_GENERATION_ATTEMPTS,
    DEFAULT_RULE_VALUE_PROBABILITY,
    DEFAULT_SEED,
)
from mex.artificial.models import RandomFieldInfo
from mex.common.fields import LITERAL_FIELDS_BY_CLASS_NAME
from mex.common.identity import get_provider
from mex.common.models import (
    ADDITIVE_MODEL_CLASSES_BY_NAME,
    EXTRACTED_MODEL_CLASSES_BY_NAME,
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    PREVENTIVE_MODEL_CLASSES_BY_NAME,
    RULE_SET_RESPONSE_CLASSES_BY_NAME,
    SUBTRACTIVE_MODEL_CLASSES_BY_NAME,
    AnyAdditiveModel,
    AnyExtractedModel,
    AnyPreventiveModel,
    AnyRuleModel,
    AnyRuleSetResponse,
    AnySubtractiveModel,
    ExtractedPrimarySource,
)
from mex.common.transform import ensure_postfix, ensure_prefix
from mex.common.types import (
    TEMPORAL_ENTITY_FORMATS_BY_PRECISION,
    UTC,
    AnyMergedIdentifier,
    Identifier,
    Link,
    LinkLanguage,
    TemporalEntity,
    TemporalEntityPrecision,
    Text,
)
from mex.common.utils import ensure_list


class BuilderProvider(PythonFakerProvider):
    """Faker provider that deals with interpreting pydantic model fields."""

    def min_max_for_field(self, field: FieldInfo) -> tuple[int, int]:
        """Return a min and max item count for a field."""
        if get_origin(field.annotation) is list:
            # calculate the item counts based on field annotations
            min_items = 0
            max_items = self.random_element(  # set probabilities for max items
                OrderedDict({1: 0.42, 2: 0.28, 3: 0.16, 4: 0.08, 5: 0.04, 10: 0.02})
            )
            if min_lengths := [x for x in field.metadata if isinstance(x, MinLen)]:
                min_items = min_lengths[0].min_length
                max_items += min_items  # the max should be higher than the min
            if max_lengths := [x for x in field.metadata if isinstance(x, MaxLen)]:
                max_items = max_lengths[0].max_length
        else:
            # required fields need at least 1 item, optional fields 0
            min_items = int(field.is_required())
            # a list with length 1 will be unpacked by `fix_listyness`
            max_items = 1
        return min_items, max_items

    def get_random_field_info(self, field: FieldInfo) -> RandomFieldInfo:
        """Randomly pick a matching type and patterns for a given field."""
        # determine field type and unpack unions, lists, and other types with args
        if args := get_args(field.annotation):
            # mixed types are not yet supported
            return self.random_element(
                [
                    self.get_random_field_info(field.from_annotation(type_))
                    for type_ in args
                    if type_ is not type(None)
                ]
            )
        return RandomFieldInfo(
            inner_type=field.annotation,
            numerify_patterns=[
                re.sub(r"[0-9]", "#", e)
                for e in ensure_list(field.examples)
                if isinstance(e, str) and not any(char in e for char in "#%$!@")
            ],
            regex_patterns=[m.pattern for m in field.metadata if hasattr(m, "pattern")],
            examples=ensure_list(field.examples),
        )

    def field_value_factory(  # noqa: PLR0911
        self,
        field_info: RandomFieldInfo,
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
    ) -> Callable[[], Any]:
        """Get a factory for creating a single value for the given field."""
        if field_info.regex_patterns and field_info.numerify_patterns:
            return partial(
                self.generator.numerify_patterns,
                field_info.numerify_patterns,
                field_info.regex_patterns,
            )
        if field_info.regex_patterns and field_info.examples:
            return partial(self.random_element, field_info.examples)
        if issubclass(field_info.inner_type, Identifier):
            return partial(
                self.generator.reference,
                field_info.inner_type,
                ids_by_type,
            )
        if issubclass(field_info.inner_type, Link):
            return partial(self.generator.link)
        if issubclass(field_info.inner_type, Text):
            return partial(self.generator.text_object)
        if issubclass(field_info.inner_type, TemporalEntity):
            return partial(
                self.generator.temporal_entity,
                field_info.inner_type.ALLOWED_PRECISION_LEVELS,
            )
        if issubclass(field_info.inner_type, Enum):
            return partial(self.random_element, field_info.inner_type)
        if issubclass(field_info.inner_type, str):
            return partial(self.generator.text_string)
        if issubclass(field_info.inner_type, int):
            return partial(self.generator.random_int)
        msg = f"Cannot create fake data for {field_info.inner_type}"
        raise RuntimeError(msg)

    def field_value(
        self,
        field: FieldInfo,
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
    ) -> Sequence[Any]:
        """Get a list of artificial values for the given field and identity."""
        field_info = self.get_random_field_info(field)
        factory = self.field_value_factory(field_info, ids_by_type)
        return list(
            dict.fromkeys(
                value
                for _ in range(self.pyint(*self.min_max_for_field(field)))
                if (value := factory()) is not None
            )
        )

    def extracted_item(
        self,
        stem_types: Sequence[str],
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
        *,
        _attempts_left: int = DEFAULT_GENERATION_ATTEMPTS,
    ) -> AnyExtractedModel:
        """Generate a single extracted item from the given stem types."""
        stem_type = self.random_element(stem_types)
        entity_type = ensure_prefix(stem_type, "Extracted")
        model_class = EXTRACTED_MODEL_CLASSES_BY_NAME[entity_type]
        # manually set provenance fields
        had_primary_source: AnyMergedIdentifier = self.random_element(
            ids_by_type[ExtractedPrimarySource.stemType]
        )
        index = len(ids_by_type[model_class.stemType])
        raw_data: dict[str, Any] = {
            "hadPrimarySource": had_primary_source,
            "identifierInPrimarySource": f"{model_class.stemType}_{index}",
            "entityType": entity_type,
        }
        # dynamically populate all other fields
        for name in sorted(model_class.model_fields):
            if name not in raw_data:
                raw_data[name] = self.field_value(
                    model_class.model_fields[name], ids_by_type
                )
        try:
            model = model_class.model_validate(raw_data)
        except ValidationError:
            if _attempts_left > 0:  # if again you don't succeed, try, try again
                return self.extracted_item(
                    stem_types, ids_by_type, _attempts_left=_attempts_left - 1
                )
            raise
        return model

    def additive_rule(
        self,
        stem_type: str,
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
        *,
        value_probability: float = DEFAULT_RULE_VALUE_PROBABILITY,
    ) -> AnyAdditiveModel:
        """Generate an artificial additive rule."""
        class_name = ensure_prefix(stem_type, "Additive")
        additive_class = ADDITIVE_MODEL_CLASSES_BY_NAME[class_name]
        raw_data = {
            name: self.field_value(additive_class.model_fields[name], ids_by_type)
            for name in sorted(additive_class.model_fields)
            if name not in LITERAL_FIELDS_BY_CLASS_NAME[class_name]
            and self.random_int() < value_probability * 1e4
        }
        return additive_class.model_validate(raw_data)

    def subtractive_rule(
        self,
        extracted_item: AnyExtractedModel,
        *,
        value_probability: float = DEFAULT_RULE_VALUE_PROBABILITY,
    ) -> AnySubtractiveModel:
        """Generate an artificial subtractive rule."""
        class_name = ensure_prefix(extracted_item.stemType, "Subtractive")
        subtractive_class = SUBTRACTIVE_MODEL_CLASSES_BY_NAME[class_name]
        raw_data = {
            name: self.random_sample(
                extracted_values,
                length=self.pyint(
                    1,
                    min(
                        len(extracted_values),
                        self.min_max_for_field(subtractive_class.model_fields[name])[1],
                    ),
                ),
            )
            for name in sorted(subtractive_class.model_fields)
            if name not in LITERAL_FIELDS_BY_CLASS_NAME[class_name]
            and self.random_int() < value_probability * 1e4
            and (extracted_values := ensure_list(getattr(extracted_item, name)))
        }
        return subtractive_class.model_validate(raw_data)

    def preventive_rule(
        self,
        extracted_item: AnyExtractedModel,
        *,
        value_probability: float = DEFAULT_RULE_VALUE_PROBABILITY,
    ) -> AnyPreventiveModel:
        """Generate an artificial preventive rule."""
        class_name = ensure_prefix(extracted_item.stemType, "Preventive")
        preventive_class = PREVENTIVE_MODEL_CLASSES_BY_NAME[class_name]
        weighted_options = OrderedDict(
            {
                (): 1 - value_probability,
                (extracted_item.hadPrimarySource,): value_probability,
            }
        )
        raw_data = {
            name: list(self.random_element(weighted_options))
            for name in preventive_class.model_fields
            if name not in LITERAL_FIELDS_BY_CLASS_NAME[class_name]
        }
        return preventive_class.model_validate(raw_data)

    def standalone_rule_set(
        self,
        stem_types: Sequence[str],
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
        identifier_seed: int = DEFAULT_SEED,
        *,
        value_probability: float = DEFAULT_RULE_VALUE_PROBABILITY,
        _attempts_left: int = DEFAULT_GENERATION_ATTEMPTS,
    ) -> AnyRuleSetResponse:
        """Generate a single standalone rule-set."""
        # manually set provenance fields
        stem_type = self.random_element(stem_types)
        rule_set_class = RULE_SET_RESPONSE_CLASSES_BY_NAME[
            ensure_postfix(stem_type, "RuleSetResponse")
        ]
        identity = get_provider().assign(
            MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
            f"artificial-rule-set-{identifier_seed}",
        )
        raw_data: dict[str, Any] = {
            "stableTargetId": identity.stableTargetId,
            "entityType": ensure_postfix(stem_type, "RuleSetResponse"),
        }
        # dynamically populate additive fields
        raw_data["additive"] = self.additive_rule(
            stem_type, ids_by_type, value_probability=value_probability
        )
        try:
            model = rule_set_class.model_validate(raw_data)
        except ValidationError:
            if _attempts_left > 0:  # if again you don't succeed, try, try again
                return self.standalone_rule_set(
                    stem_types,
                    ids_by_type,
                    identifier_seed,
                    _attempts_left=_attempts_left - 1,
                )
            raise
        return model

    def rule_set_for_item(
        self,
        extracted_item: AnyExtractedModel,
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
        *,
        value_probability: float = DEFAULT_RULE_VALUE_PROBABILITY,
        _attempts_left: int = DEFAULT_GENERATION_ATTEMPTS,
    ) -> AnyRuleSetResponse:
        """Generate a single rule-set for the given extracted item."""
        # manually set provenance fields
        stem_type = extracted_item.stemType
        rule_set_class = RULE_SET_RESPONSE_CLASSES_BY_NAME[
            ensure_postfix(stem_type, "RuleSetResponse")
        ]
        raw_data: dict[str, Any] = {
            "stableTargetId": extracted_item.stableTargetId,
            "entityType": ensure_postfix(stem_type, "RuleSetResponse"),
        }
        # randomize which rules to populate, could be: none, some or all
        rule_names_and_factories = self.random_sample(
            [
                (
                    "additive",
                    partial(
                        self.additive_rule,
                        stem_type,
                        ids_by_type,
                        value_probability=value_probability,
                    ),
                ),
                (
                    "preventive",
                    partial(
                        self.preventive_rule,
                        extracted_item,
                        value_probability=value_probability,
                    ),
                ),
                (
                    "subtractive",
                    partial(
                        self.subtractive_rule,
                        extracted_item,
                        value_probability=value_probability,
                    ),
                ),
            ]
        )
        # dynamically populate selected rule fields
        for rule_name, factory in rule_names_and_factories:
            raw_data[rule_name] = cast("Callable[[], AnyRuleModel]", factory)()
        try:
            model = rule_set_class.model_validate(raw_data)
        except ValidationError:
            if _attempts_left > 0:  # if again you don't succeed, try, try again
                return self.rule_set_for_item(
                    extracted_item, ids_by_type, _attempts_left=_attempts_left - 1
                )
            raise
        return model


class ReferenceProvider(BaseFakerProvider):
    """Faker provider that creates references to other items."""

    def reference(
        self,
        inner_type: type[Identifier],
        ids_by_type: Mapping[str, Collection[AnyMergedIdentifier]],
    ) -> AnyMergedIdentifier | None:
        """Return random merged item identifier picked from available mapping."""
        if choices := [
            identifier
            for stem_type in re.findall(r"Merged([A-Za-z]+)Identifier", str(inner_type))
            for identifier in ids_by_type[stem_type]
        ]:
            return self.random_element(choices)
        return None


class LinkProvider(InternetFakerProvider, PythonFakerProvider):
    """Faker provider that can return links with optional title and language."""

    def link(self) -> Link:
        """Return a link with optional title and language."""
        title, language = None, None
        if self.pybool():
            title = self.domain_word().replace("-", " ").title()
            if self.pybool():
                language = self.random_element(LinkLanguage)
        return Link(url=self.url(), title=title, language=language)


class TemporalEntityProvider(PythonFakerProvider):
    """Faker provider that can return a custom TemporalEntity with random precision."""

    def temporal_entity(
        self,
        allowed_precision_levels: list[TemporalEntityPrecision],
    ) -> TemporalEntity:
        """Return a custom temporal entity with random date, time and precision."""
        return TemporalEntity(
            datetime.fromtimestamp(
                self.pyint(int(8e8), int(datetime.now(tz=UTC).timestamp())), tz=UTC
            ).strftime(
                TEMPORAL_ENTITY_FORMATS_BY_PRECISION[
                    self.random_element(allowed_precision_levels)
                ]
            )
        )


class TextProvider(PythonFakerProvider):
    """Faker provider that handles custom text related requirements."""

    def __init__(self, factory: FakerFactory, chattiness: int) -> None:
        """Configure the chattiness of generated text."""
        super().__init__(factory)
        self._chattiness = chattiness

    def text_string(self) -> str:
        """Return a randomized sequence of words as a string."""
        return " ".join(
            self.generator.word()
            for _ in range(self.pyint(1, max(int(self._chattiness / 5), 3)))
        )

    def text_object(self) -> Text:
        """Return a random text paragraph with an auto-detected language."""
        return Text(value=self.generator.paragraph(self.pyint(1, self._chattiness)))


class NumerifyPatternsProvider(PythonFakerProvider):
    """Faker provider that tries to numerify a pattern until it matches a regex."""

    def numerify_patterns(
        self,
        numerify_patterns: list[str],
        regex_patterns: list[str],
    ) -> str | None:
        """Try to numerify a pattern in 10 turns until it validates, or bail out."""
        for _ in range(10):
            numerify_pattern = self.random_element(numerify_patterns)
            regex_pattern = self.random_element(regex_patterns)
            numerified = self.numerify(numerify_pattern)
            if re.match(regex_pattern, numerified):
                return numerified
        return None



from typing import Any

from pydantic import BaseModel

from mex.common.models import AnyExtractedModel, AnyRuleSetResponse
from mex.common.types import AnyPrimitiveType


class RandomFieldInfo(BaseModel):
    """Randomized pick of matching inner type and patterns for a field."""

    inner_type: Any
    numerify_patterns: list[str] = []
    regex_patterns: list[str] = []
    examples: list[AnyPrimitiveType | dict[str, AnyPrimitiveType]] = []


class ExtractedItemAndRuleSet(BaseModel):
    """A combination of an extracted item and a rule-set (both optional)."""

    extracted_item: AnyExtractedModel | None = None
    rule_set: AnyRuleSetResponse | None = None



from mex.common.models import (
    MEX_PRIMARY_SOURCE_IDENTIFIER,
    MEX_PRIMARY_SOURCE_IDENTIFIER_IN_PRIMARY_SOURCE,
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    ExtractedPrimarySource,
)

DEFAULT_LOCALE = [
    "de_DE",
    "en_US",
]
DEFAULT_MODELS = [
    "AccessPlatform",
    "Activity",
    "BibliographicResource",
    "ContactPoint",
    "Distribution",
    "Organization",
    "OrganizationalUnit",
    "Person",
    "Resource",
    "Variable",
    "VariableGroup",
]
DEFAULT_SEED = 0
DEFAULT_CHATTINESS = 10
DEFAULT_COUNT = 100
DEFAULT_RULE_VALUE_PROBABILITY = 0.33
DEFAULT_GENERATION_ATTEMPTS = 10

MEX_PRIMARY_SOURCE = ExtractedPrimarySource.model_construct(
    hadPrimarySource=MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    identifier=MEX_PRIMARY_SOURCE_IDENTIFIER,
    identifierInPrimarySource=MEX_PRIMARY_SOURCE_IDENTIFIER_IN_PRIMARY_SOURCE,
    stableTargetId=MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
)



from typing import Annotated, Any

import pytest
from faker import Faker
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

from mex.artificial.models import RandomFieldInfo
from mex.common.types import (
    APIType,
    Link,
    MergedOrganizationIdentifier,
    MergedPrimarySourceIdentifier,
    TemporalEntity,
    TemporalEntityPrecision,
    Text,
    TextLanguage,
    YearMonth,
    YearMonthDay,
)


class DummyModel(BaseModel):
    no_min: list[str] = []
    has_min: list[bytes] = Field([], min_length=2)
    has_max: list[bytes] = Field([], max_length=5)
    is_required: int
    is_optional: bool | None = None
    is_union: float | list[float]
    is_inner_union: list[float | int] = []
    is_union_with_pattern: (
        Annotated[
            str,
            Field(
                pattern=r"^https://wikidata\.org/entity/[PQ0-9]{2,64}$",
                examples=[
                    "http://wikidata.org/entity/Q679041",
                    "http://wikidata.org/entity/P123",
                ],
            ),
        ]
        | None
    ) = None
    is_nested_pattern: list[
        Annotated[
            str,
            Field(
                pattern=r"^https://dfg\.de/foobar/[0-9]{1,64}$",
                examples=["https://dfg.de/foobar/10179"],
            ),
        ]
    ] = []


def test_builder_provider_min_max_for_field(faker: Faker) -> None:
    min_max = {
        name: faker.min_max_for_field(field)
        for name, field in DummyModel.model_fields.items()
    }
    assert min_max == {
        "no_min": (0, 4),
        "has_min": (2, 6),
        "has_max": (0, 5),
        "is_required": (1, 1),
        "is_optional": (0, 1),
        "is_union": (1, 1),
        "is_inner_union": (0, 3),
        "is_union_with_pattern": (0, 1),
        "is_nested_pattern": (0, 5),
    }


def test_builder_provider_get_random_field_info(faker: Faker) -> None:
    random_field_info = {
        name: faker.get_random_field_info(field)
        for name, field in DummyModel.model_fields.items()
    }
    assert random_field_info == {
        "no_min": RandomFieldInfo(inner_type=str),
        "has_min": RandomFieldInfo(inner_type=bytes),
        "has_max": RandomFieldInfo(inner_type=bytes),
        "is_required": RandomFieldInfo(inner_type=int),
        "is_optional": RandomFieldInfo(inner_type=bool),
        "is_union": RandomFieldInfo(inner_type=float),
        "is_inner_union": RandomFieldInfo(inner_type=int),
        "is_union_with_pattern": RandomFieldInfo(
            inner_type=str,
            numerify_patterns=[
                "http://wikidata.org/entity/Q######",
                "http://wikidata.org/entity/P###",
            ],
            regex_patterns=[r"^https://wikidata\.org/entity/[PQ0-9]{2,64}$"],
            examples=[
                "http://wikidata.org/entity/Q679041",
                "http://wikidata.org/entity/P123",
            ],
        ),
        "is_nested_pattern": RandomFieldInfo(
            inner_type=str,
            numerify_patterns=["https://dfg.de/foobar/#####"],
            regex_patterns=[r"^https://dfg\.de/foobar/[0-9]{1,64}$"],
            examples=["https://dfg.de/foobar/10179"],
        ),
    }


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        (
            Annotated[
                str,
                Field(
                    pattern=r"^https://foo\.example/[a-z0-9]{9}$",
                    examples=["https://foo.example/01k5qnb77"],
                ),
            ],
            ["https://foo.example/48k7qnb64"],
        ),
        (
            Annotated[
                str,
                Field(
                    examples=["info@rki.de"],
                    pattern="^[^@ \\t\\r\\n]+@[^@ \\t\\r\\n]+\\.[^@ \\t\\r\\n]+$",
                    json_schema_extra={"format": "email"},
                ),
            ],
            ["info@rki.de"],
        ),
        (
            MergedPrimarySourceIdentifier,
            ["PrimarySource00000007"],
        ),
        (
            Link,
            [
                Link(url="https://www.sheppard-tucker.com/"),
            ],
        ),
        (
            Text,
            [
                Text(
                    value="Whole magazine truth stop whose. On traditional measure "
                    "example sense peace.",
                    language=TextLanguage.EN,
                )
            ],
        ),
        (
            YearMonthDay | YearMonth,
            [TemporalEntity("1996-09")],
        ),
        (
            APIType,
            [APIType["OTHER"]],
        ),
        (
            str,
            ["amount event"],
        ),
        (
            int,
            [6890],
        ),
    ],
)
def test_builder_provider_field_value(
    faker: Faker,
    ids_by_type: dict[str, list[str]],
    annotation: Any,  # noqa: ANN401
    expected: Any,  # noqa: ANN401
) -> None:
    field = FieldInfo.from_annotation(annotation)
    assert faker.field_value(field, ids_by_type) == expected


def test_builder_provider_field_value_error(faker: Faker) -> None:
    field = FieldInfo.from_annotation(object)

    with pytest.raises(RuntimeError, match="Cannot create fake data"):
        faker.field_value(field, {})


def test_builder_provider_extracted_item(
    faker: Faker, ids_by_type: dict[str, list[str]]
) -> None:
    extracted_item = faker.extracted_item(
        ["ContactPoint"],
        ids_by_type,
    )
    assert extracted_item.model_dump(exclude_defaults=True) == {
        "hadPrimarySource": "PrimarySource00000007",
        "identifierInPrimarySource": "ContactPoint_8",
        "email": ["info@rki.de"],
        "identifier": "fVWC01006cQWIynNJ526b3",
        "stableTargetId": "htOot14HcSLqexpYQsMKwj",
    }


def test_builder_provider_additive_rule(
    faker: Faker, ids_by_type: dict[str, list[str]]
) -> None:
    rule = faker.additive_rule(
        "Person",
        ids_by_type,
        value_probability=1,
    )
    assert rule.model_dump(exclude_defaults=True) == {
        "affiliation": ["Organization000000008"],
        "email": ["info@rki.de"],
        "familyName": ["season take play"],
        "isniId": ["https://isni.org/isni/7840801609753513"],
    }


def test_builder_provider_subtractive_rule(
    faker: Faker, ids_by_type: dict[str, list[str]]
) -> None:
    extracted_item = faker.extracted_item(["Person"], ids_by_type)
    rule = faker.subtractive_rule(
        extracted_item,
        value_probability=1,
    )
    assert rule.model_dump(exclude_defaults=True) == {
        "affiliation": [
            "Organization000000008",
            "Organization000000005",
            "Organization000000007",
        ],
        "email": ["info@rki.de"],
        "fullName": ["show"],
        "givenName": ["me level tree"],
        "memberOf": ["OrganizationalUnit008", "OrganizationalUnit001"],
        "orcidId": ["https://orcid.org/8711-5871-4841-858X"],
    }


def test_builder_provider_preventive_rule(
    faker: Faker, ids_by_type: dict[str, list[str]]
) -> None:
    extracted_item = faker.extracted_item(["Person"], ids_by_type)
    rule = faker.preventive_rule(
        extracted_item,
        value_probability=0.75,
    )
    assert rule.model_dump(exclude_defaults=True) == {
        "email": ["PrimarySource00000007"],
        "familyName": ["PrimarySource00000007"],
        "givenName": ["PrimarySource00000007"],
        "isniId": ["PrimarySource00000007"],
        "orcidId": ["PrimarySource00000007"],
    }


def test_builder_provider_standalone_rule_set(
    faker: Faker, ids_by_type: dict[str, list[str]]
) -> None:
    rule_set = faker.standalone_rule_set(
        ["Person"],
        ids_by_type,
        identifier_seed=42,
        value_probability=0.75,
    )
    assert rule_set.model_dump(exclude_defaults=True) == {
        "additive": {
            "affiliation": [
                "Organization000000007",
                "Organization000000005",
                "Organization000000008",
            ],
            "email": ["info@rki.de"],
            "familyName": ["season take play", "force"],
            "fullName": [
                "opportunity blood",
                "leader four",
                "push while democratic",
                "me level tree",
                "ever assume support",
            ],
            "isniId": [
                "https://isni.org/isni/9753513933287115",
                "https://isni.org/isni/1484185839894719",
                "https://isni.org/isni/9342320947112201",
                "https://isni.org/isni/8483396947751591",
                "https://isni.org/isni/3304135256012309",
                "https://isni.org/isni/0139916151090321",
            ],
        },
        "stableTargetId": "b3wrW78ASMOyBrXF6ikQAO",
    }


def test_reference_provider(faker: Faker, ids_by_type: dict[str, list[str]]) -> None:
    inner_type = list[MergedPrimarySourceIdentifier | MergedOrganizationIdentifier]
    assert faker.reference(inner_type, ids_by_type) == "Organization000000005"
    assert faker.reference(dict[str, str], ids_by_type) is None


def test_link_provider(faker: Faker) -> None:
    assert faker.link() == Link(
        language=None, title="Williams Sheppard", url="https://howard-snow.com/"
    )


def test_temporal_entity_provider(faker: Faker) -> None:
    assert faker.temporal_entity([TemporalEntityPrecision.DAY]) == TemporalEntity(
        "2024-01-31"
    )


def test_text_provider_string(faker: Faker) -> None:
    assert faker.text_string() == "son voice"


def test_text_provider_text(faker: Faker) -> None:
    assert faker.text_object() == Text(
        value="Else memory if. Whose group through despite cause. Sense peace economy "
        "travel. Total financial role together range line beyond its.",
        language=TextLanguage.EN,
    )



from pathlib import Path
from typing import Annotated

import typer

from mex.artificial.constants import (
    DEFAULT_CHATTINESS,
    DEFAULT_COUNT,
    DEFAULT_LOCALE,
    DEFAULT_MODELS,
    DEFAULT_SEED,
)
from mex.artificial.helpers import generate_artificial_merged_items, write_merged_items
from mex.common.logging import logger


def artificial(  # noqa: PLR0913
    count: Annotated[
        int,
        typer.Option(
            help="Number of artificial items to approximately create.",
            min=1,
            max=int(10e6 - 1),
        ),
    ] = DEFAULT_COUNT,
    chattiness: Annotated[
        int,
        typer.Option(
            help="Maximum number of words to approximately produce for text fields.",
            min=2,
            max=100,
        ),
    ] = DEFAULT_CHATTINESS,
    seed: Annotated[
        int,
        typer.Option(
            help="The seed value for faker randomness.",
        ),
    ] = DEFAULT_SEED,
    locale: Annotated[
        list[str] | None,
        typer.Option(
            help="The locale to use for faker.",
            show_default=DEFAULT_LOCALE,
        ),
    ] = None,
    models: Annotated[
        list[str] | None,
        typer.Option(
            help="The names of models to use for faker.",
            show_default=DEFAULT_MODELS,
        ),
    ] = None,
    path: Annotated[
        Path | None,
        typer.Option(
            help="Where to write the resulting `publisher.ndjson` file.",
            show_default=str(Path.cwd()),
            file_okay=False,
            dir_okay=True,
            writable=True,
            resolve_path=True,
        ),
    ] = None,
) -> None:  # pragma: no cover
    """Generate merged artificial items."""
    logger.info("starting artificial data generation")
    path = path or Path.cwd()
    locale = locale or DEFAULT_LOCALE
    models = models or DEFAULT_MODELS
    item_gen = generate_artificial_merged_items(locale, seed, chattiness, models)
    write_merged_items(item_gen, count, path)
    logger.info("artificial data generation done")


def main() -> None:  # pragma: no cover
    """Wrap entrypoint in typer."""
    typer.run(artificial)



import json
from collections import OrderedDict, defaultdict
from collections.abc import Generator, Iterable, Mapping, Sequence
from itertools import count, islice
from os import PathLike
from pathlib import Path
from typing import cast

from faker import Faker
from faker.typing import SeedType
from rich.progress import track

from mex.artificial.constants import (
    DEFAULT_CHATTINESS,
    DEFAULT_COUNT,
    DEFAULT_LOCALE,
    DEFAULT_MODELS,
    DEFAULT_SEED,
    MEX_PRIMARY_SOURCE,
)
from mex.artificial.models import ExtractedItemAndRuleSet
from mex.artificial.provider import (
    BuilderProvider,
    LinkProvider,
    NumerifyPatternsProvider,
    ReferenceProvider,
    TemporalEntityProvider,
    TextProvider,
)
from mex.artificial.types import LocaleType
from mex.common.merged.main import create_merged_item
from mex.common.models import AnyExtractedModel, AnyMergedModel
from mex.common.transform import MExEncoder
from mex.common.types import AnyMergedIdentifier, Validation


def create_faker(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
) -> Faker:
    """Create and initialize a new faker instance with the given settings."""
    Faker.seed(seed)
    faker = Faker(locale=locale)
    for factory in faker.factories:
        factory.add_provider(ReferenceProvider(factory))
        factory.add_provider(LinkProvider(factory))
        factory.add_provider(NumerifyPatternsProvider(factory))
        factory.add_provider(BuilderProvider(factory))
        factory.add_provider(TextProvider(factory, chattiness))
        factory.add_provider(TemporalEntityProvider(factory))
    return faker


def generate_artificial_extracted_items(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
    stem_types: Sequence[str] = DEFAULT_MODELS,
) -> Generator[AnyExtractedModel, None, None]:
    """Infinitely generate extracted items for the given settings."""
    faker = create_faker(locale, seed, chattiness)
    ids_by_type: Mapping[str, OrderedDict[AnyMergedIdentifier, None]] = defaultdict(
        OrderedDict,
        {
            MEX_PRIMARY_SOURCE.stemType: OrderedDict(
                {MEX_PRIMARY_SOURCE.stableTargetId: None}
            )
        },
    )
    yield MEX_PRIMARY_SOURCE
    while True:
        item = cast("AnyExtractedModel", faker.extracted_item(stem_types, ids_by_type))
        ids_by_type[item.stemType][item.stableTargetId] = None
        yield item


def create_artificial_extracted_items(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
    stem_types: Sequence[str] = DEFAULT_MODELS,
    count: int = DEFAULT_COUNT,
) -> list[AnyExtractedModel]:
    """Create a list of extracted items for the given settings."""
    return list(
        islice(
            generate_artificial_extracted_items(locale, seed, chattiness, stem_types),
            count,
        )
    )


def generate_artificial_items_and_rule_sets(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
    stem_types: Sequence[str] = DEFAULT_MODELS,
) -> Generator[ExtractedItemAndRuleSet, None, None]:
    """Infinitely generate artificial extracted items and rule-sets."""
    faker = create_faker(locale, seed, chattiness)
    ids_by_type: Mapping[str, OrderedDict[AnyMergedIdentifier, None]] = defaultdict(
        OrderedDict,
        {
            MEX_PRIMARY_SOURCE.stemType: OrderedDict(
                {MEX_PRIMARY_SOURCE.stableTargetId: None}
            )
        },
    )
    yield ExtractedItemAndRuleSet(extracted_item=MEX_PRIMARY_SOURCE)
    for index in count():
        match faker.random_int(0, 2):
            case 0:  # only extracted item
                item = faker.extracted_item(stem_types, ids_by_type)
                ids_by_type[item.stemType][item.stableTargetId] = None
                yield ExtractedItemAndRuleSet(extracted_item=item)
            case 1:  # only rule set
                rule_set = faker.standalone_rule_set(stem_types, ids_by_type, index)
                ids_by_type[rule_set.stemType][rule_set.stableTargetId] = None
                yield ExtractedItemAndRuleSet(rule_set=rule_set)
            case 2:  # extracted item and rule set
                item = faker.extracted_item(stem_types, ids_by_type)
                rule_set = faker.rule_set_for_item(item, ids_by_type)
                ids_by_type[item.stemType][item.stableTargetId] = None
                yield ExtractedItemAndRuleSet(extracted_item=item, rule_set=rule_set)


def create_artificial_items_and_rule_sets(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
    stem_types: Sequence[str] = DEFAULT_MODELS,
    count: int = DEFAULT_COUNT,
) -> list[ExtractedItemAndRuleSet]:
    """Create a list of artificial extracted items and rule-sets."""
    return list(
        islice(
            generate_artificial_items_and_rule_sets(
                locale, seed, chattiness, stem_types
            ),
            count,
        )
    )


def generate_artificial_merged_items(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
    stem_types: Sequence[str] = DEFAULT_MODELS,
) -> Generator[AnyMergedModel, None, None]:
    """Infinitely generate merged items for the given settings."""
    for item_combination in generate_artificial_items_and_rule_sets(
        locale, seed, chattiness, stem_types
    ):
        identifier = next(
            i.stableTargetId
            for i in (item_combination.extracted_item, item_combination.rule_set)
            if i
        )
        if merged_item := create_merged_item(
            identifier,
            [item_combination.extracted_item]
            if item_combination.extracted_item
            else [],
            item_combination.rule_set,
            validation=Validation.IGNORE,
        ):
            yield merged_item


def create_artificial_merged_items(
    locale: LocaleType = DEFAULT_LOCALE,
    seed: SeedType = DEFAULT_SEED,
    chattiness: int = DEFAULT_CHATTINESS,
    stem_types: Sequence[str] = DEFAULT_MODELS,
    count: int = DEFAULT_COUNT,
) -> list[AnyMergedModel]:
    """Create a list of merged items for the given settings."""
    return list(
        islice(
            generate_artificial_merged_items(locale, seed, chattiness, stem_types),
            count,
        )
    )


def write_merged_items(
    items: Iterable[AnyMergedModel],
    count: int = DEFAULT_COUNT,
    out_path: PathLike[str] | None = None,
) -> None:
    """Write the desired number of items from the incoming stream to an NDJSON file."""
    file_path = Path(out_path or Path.cwd()) / "publisher.ndjson"
    with file_path.open("w", encoding="utf-8") as fh:
        for item in track(islice(items, count), total=count, description="working..."):
            line = json.dumps(item, ensure_ascii=False, sort_keys=True, cls=MExEncoder)
            fh.write(f"{line}\n")
